# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.



import copy
import re

from nbdime import patch
from nbdime.diff_format import op_patch
from nbdime.merging.decisions import (
    apply_decisions, ensure_common_path, MergeDecision)

from nbdime import diff


def has_merge_conflicts(decisions):
    "Return whether there are conflicting entries or not."
    return any(item.conflict for item in decisions)


_r_is_int = re.compile("^[0-9]+$")


def is_int(n):
    return bool(_r_is_int.match(n))


def pick_merge_decision(base, dec):
    if dec.action is None or dec.action == "base":
        di = None
    elif dec.action == "local" or dec.action == "either":
        di = dec.local_diff
    elif dec.action == "remote":
        di = dec.remote_diff
    elif dec.action == "custom":
        di = dec.custom_diff
    else:
        raise ValueError("Unknown action {}".format(dec.action))

    if di is not None:
        # Parse common path
        keys = [k for k in dec.common_path.split("/") if k != ""]
        sub = base
        for k in keys:
            if isinstance(sub, list):
                k = int(k)
            sub = sub[k]
        # "/cells" -> sub = base[cells], sub is a list
        # patch

        # Add patch entries
        base_diff = di
        for k in reversed(keys):
            if is_int(k):
                k = int(k)
            base_diff = op_patch(k, base_diff)
        # Apply patch
        base = patch(base, base_diff)

    return base


def gather_merge_decisions(base, decisions):
    assert not has_merge_conflicts(decisions)
    for dec in decisions:
        base = pick_merge_decision(base, dec)
    return base


def create_decision_item(action=None, common_path="", conflict=False,
                         local_diff=None, remote_diff=None, custom_diff=None):

    # Some parameter validation here, but don't need to duplicate json schema
    if action is None:
        pass
    elif action == "local":
        assert local_diff
    elif action == "remote":
        assert remote_diff
    elif action == "custom":
        assert custom_diff
    elif action in ("either", "local_then_remote", "remote_then_local"):
        assert local_diff
        assert remote_diff
    else:
        pass

    item = MergeDecision({
            "action": action,
            "common_path": common_path,
            "conflict": conflict,
            "custom_diff": custom_diff,
            "local_diff": local_diff,
            "remote_diff": remote_diff,
        })
    return item


def _example_decisions():
    decisions = [
        create_decision_item(
            action="local",
            local_diff=""
            )
        ]
    return decisions


def test_apply_merge_empty():
    decisions = []
    base = {"hello": "world"}
    assert base == apply_decisions(base, decisions)


def test_apply_merge_on_dicts():
    base = {
        "metadata": {
            "a": {"ting": 123},
            "b": {"tang": 456}
        }
    }

    local = copy.deepcopy(base)
    local["metadata"]["a"]["ting"] += 1

    remote = copy.deepcopy(base)
    remote["metadata"]["a"]["ting"] -= 1

    bld = diff(base, local)
    brd = diff(base, remote)

    path, (bld, brd) = ensure_common_path((), [bld, brd])

    merge_decisions = [
        create_decision_item(
            action="remote",
            common_path=path,
            local_diff=bld,
            remote_diff=brd)
    ]

    assert remote == apply_decisions(base, merge_decisions)

# merge decisions with common path "cells" can modify cells/* indices
# merge decisions with common path "cells/*" only edit exactly one of the cells/* objects
# applying cells/* before cells means editing first, no indices modified, then moving things around



# -*- coding: utf-8 -*-

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.


def test_pkg_import_no_webapp():
    # Test that importing nbdime does not import webapp
    # as this is an expensive import
    import nbdime
    import sys
    assert 'ndime.webapp' not in sys.modules



# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import logging


class NBDiffFormatError(ValueError):
    pass


def init_logging(level=logging.INFO):
    """Sets up logging for nbdime entry points.

    Call this in all entry points (if __name__ == "__main__").
    Sets the log level for all nbdime loggers to `level`,
    unless `level` is given as `None`.
    """
    format = '[%(levelname)1.1s %(module)s:%(lineno)d] %(message)s'
    logging.basicConfig(format=format, level=level)
    logging.captureWarnings(True)


def set_nbdime_log_level(level, set_main=True):
    """Set a log level for nbdime loggers"""
    logger.setLevel(level)
    if set_main:
        _baseLogger = logging.getLogger()
        _baseLogger.setLevel(level)


logger = logging.getLogger('nbdime')

debug = logger.debug
info = logger.info
warning = logger.warning
error = logger.error
exception = logger.exception
critical = logger.critical




import os

from jupyter_core.paths import jupyter_config_path

from traitlets import Unicode, Enum, Integer, Bool, HasTraits, Dict, TraitError
from traitlets.config.loader import JSONFileConfigLoader, ConfigFileNotFound

from .merging.notebooks import (
    cli_conflict_strategies, cli_conflict_strategies_input, cli_conflict_strategies_output)




class NbdimeConfigurable(HasTraits):

    def configured_traits(self, cls):
        traits = cls.class_own_traits(config=True)
        c = {}
        for name, _ in traits.items():
            c[name] = getattr(self, name)
        return c


_config_cache = {}
def config_instance(cls):
    if cls in _config_cache:
        return _config_cache[cls]
    instance = _config_cache[cls] = cls()
    return instance


def _load_config_files(basefilename, path=None):
    """Load config files (json) by filename and path.

    yield each config object in turn.
    """

    if not isinstance(path, list):
        path = [path]
    for path in path[::-1]:
        # path list is in descending priority order, so load files backwards:
        loader = JSONFileConfigLoader(basefilename+'.json', path=path)
        config = None
        try:
            config = loader.load_config()
        except ConfigFileNotFound:
            pass
        if config:
            yield config


def recursive_update(target, new, include_none):
    """Recursively update one dictionary using another.

    None values will delete their keys.
    """
    for k, v in new.items():
        if isinstance(v, dict):
            if k not in target:
                target[k] = {}
            recursive_update(target[k], v, include_none)
            if not include_none and not target[k]:
                # Prune empty subdicts
                del target[k]

        elif not include_none and v is None:
            target.pop(k, None)

        else:
            target[k] = v


def build_config(entrypoint, include_none=False):
    if entrypoint not in entrypoint_configurables:
        raise ValueError('Config for entrypoint name %r is not defined! Accepted values are %r.' % (
            entrypoint, list(entrypoint_configurables.keys())
        ))

    # Get config from disk:
    disk_config = {}
    path = jupyter_config_path()
    path.insert(0, os.getcwd())
    for c in _load_config_files('nbdime_config', path=path):
        recursive_update(disk_config, c, include_none)

    config = {}
    configurable = entrypoint_configurables[entrypoint]
    for c in reversed(configurable.mro()):
        if issubclass(c, NbdimeConfigurable):
            recursive_update(config, config_instance(c).configured_traits(c), include_none)
            if (c.__name__ in disk_config):
                recursive_update(config, disk_config[c.__name__], include_none)

    return config


def get_defaults_for_argparse(entrypoint):
    return build_config(entrypoint)


class Global(NbdimeConfigurable):

    log_level = Enum(
        ('DEBUG', 'INFO', 'WARN', 'ERROR', 'CRITICAL'),
        'INFO',
        help="Set the log level by name.",
    ).tag(config=True)


class Web(NbdimeConfigurable):

    port = Integer(
        0,
        help="specify the port you want the server to run on. Default is 0 (random).",
    ).tag(config=True)

    ip = Unicode(
        '127.0.0.1',
        help="specify the interface to listen to for the web server. "
        "NOTE: Setting this to anything other than 127.0.0.1/localhost "
        "might comprimise the security of your computer. Use with care!",
    ).tag(config=True)

    base_url = Unicode(
        '/', help="The base URL prefix under which to run the web app",
    ).tag(config=True)

    browser = Unicode(
        None,
        allow_none=True,
        help="specify the browser to use, to override the system default.",
    ).tag(config=True)

    persist = Bool(
        False, help="prevent server shutting down on remote close request "
                    "(when these would normally be supported).",
    ).tag(config=True)

    workdirectory = Unicode(
        default_value=os.path.abspath(os.path.curdir),
        help="specify the working directory you want "
             "the server to run from. Default is the "
             "actual cwd at program start.",
    ).tag(config=True)


class WebTool(Web):
    pass



class IgnoreConfig(Dict):

    def validate_elements(self, obj, value):
        value = super(IgnoreConfig, self).validate_elements(obj, value)
        for k, v in value.items():
            if not k.startswith('/'):
                raise TraitError('ignore config paths need to start with `/`')
            if not (v in (True, False) or
                    (isinstance(v, (tuple, list, set)) and
                     all(isinstance(i, str) for i in v)
                    )):
                raise TraitError('ignore config value needs to be True, False or a list of strings')
        return self.klass(value)


class _Ignorables(NbdimeConfigurable):

    sources = Bool(
        None,
        allow_none=True,
        help="process/ignore sources.",
    ).tag(config=True)

    outputs = Bool(
        None,
        allow_none=True,
        help="process/ignore outputs.",
    ).tag(config=True)

    metadata = Bool(
        None,
        allow_none=True,
        help="process/ignore metadata.",
    ).tag(config=True)

    id = Bool(
        None,
        allow_none=True,
        help="process/ignore identifiers.",
    ).tag(config=True)

    attachments = Bool(
        None,
        allow_none=True,
        help="process/ignore attachments.",
    ).tag(config=True)

    details = Bool(
        None,
        allow_none=True,
        help="process/ignore details not covered by other options.",
    ).tag(config=True)

    Ignore = IgnoreConfig(
        default_value={},
        help="a custom ignore config"
    ).tag(config=True)


class Show(_Ignorables):
    pass


class _Diffing(_Ignorables):

    color_words = Bool(
        False,
        help=("whether to pass the --color-words flag to any internal calls "
              "to git diff"),
    ).tag(config=True)


class Diff(_Diffing):
    pass


class Merge(_Diffing):

    merge_strategy = Enum(
        cli_conflict_strategies,
        'inline',
        help="Specify the merge strategy to use.",
    ).tag(config=True)

    input_strategy = Enum(
        cli_conflict_strategies_input,
        None,
        allow_none=True,
        help="Specify the merge strategy to use for inputs "
             "(overrides 'merge-strategy' for inputs).",
    ).tag(config=True)

    output_strategy = Enum(
        cli_conflict_strategies_output,
        None,
        allow_none=True,
        help="Specify the merge strategy to use for outputs "
             "(overrides 'merge-strategy' for outputs).",
    ).tag(config=True)

    ignore_transients = Bool(
        True,
        help="Disallow deletion of transient data such as outputs and "
             "execution counts in order to resolve conflicts.",
    ).tag(config=True)


class GitDiff(Diff):
    use_filter = Bool(
        False,
        help="if enabled, will apply any configured git filter to the remote file "
             "prior to diffing."
    )

class GitMerge(Merge):
    pass


class NbDiff(GitDiff):
    pass

class NbDiffWeb(Web, GitDiff):
    pass

class NbMerge(Merge):
    pass

class NbMergeWeb(Web, Merge):

    show_base = Bool(
        True,
        help="Whether to show the base version (4-panels) or not (3-panels).",
    ).tag(config=True)

class NbShow(Show):
    pass

class Server(Web):

    port = Integer(
        8888,
        help="specify the port you want the server to run on. Default is 8888.",
    ).tag(config=True)


class NbDiffDriver(GitDiff):
    pass

class NbDiffTool(GitDiff, WebTool):
    pass

class NbMergeDriver(GitMerge):
    pass

class NbMergeTool(GitMerge, WebTool):
    pass

class Extension(GitDiff):
    pass


entrypoint_configurables = {
    'nbdiff': NbDiff,
    'nbdiff-web': NbDiffWeb,
    'nbmerge': NbMerge,
    'nbmerge-web': NbMergeWeb,
    'nbshow': NbShow,
    'server': Server,
    'extension': Extension,
    'git-nbdiffdriver': NbDiffDriver,
    'git-nbdifftool': NbDiffTool,
    'git-nbmergedriver': NbMergeDriver,
    'git-nbmergetool': NbMergeTool,
}


class Namespace(object):
    def __init__(self, adict):
        self.__dict__.update(adict)




import io
from subprocess import check_output, STDOUT, CalledProcessError

from io import StringIO

from nbdime.utils import EXPLICIT_MISSING_FILE


class NamedStringIO(StringIO):
    name = ''


def interrogate_filter(path):
    """Check whether a filter git attribute is set for path.

    Returns None if no valid filter attribute could be found.
    """
    # Ask for the filter attributes of file on path (-z = null-terminated fields)
    try:
        spec = check_output(['git', 'check-attr', '-z', 'filter', path])
    except CalledProcessError:
        return None
    try:
        path_list, attr, info = [s.decode('utf8', 'replace') for s in spec.split(b'\x00')[:3]]
    except ValueError:
        # For older versions of git, the `-z` flag is unsupported
        # This will then raise a ValueError when trying to unpack
        return None
    if attr != 'filter' or path != path_list:
        raise ValueError(
            'Unexpected output from git check-attr. ' +
            ('Expected to start with "%s\x00filter", ' % path) +
            ('started with "%s\x00%s"' % (path_list, attr))
        )
    if not info or info in ('unspecified', 'set', 'unset'):
        return None
    # We have a valid attribute
    return info


def get_clean_filter_cmd(filter_attr):
    """Given a filter attribute, look up its driver in git config.

    Returns None if noe valid config could be found.
    """
    try:
        spec = check_output([
            'git', 'config', '--get', '-z', 'filter.%s.clean' % filter_attr
        ])
        return spec.split(b'\x00')[0].decode('utf8', 'replace') or None
    except CalledProcessError:
        return None


def apply_possible_filter(git_path, path=None):
    """Apply any configured git filters to path.

    Returns the original remote path if no filter is configured,
    or a StringIO instance with the filtered content if a filter
    should be applied.
    """
    if path is None:
        path = git_path

    if path == EXPLICIT_MISSING_FILE:
        return path

    filter_attr = interrogate_filter(git_path)
    if not filter_attr:
        return path
    filter_cmd = get_clean_filter_cmd(filter_attr)
    if not filter_cmd:
        return path

    # Apply filter and pipe to a string buffer
    with io.open(path, 'r', encoding="utf8") as f:
        output = check_output(
            filter_cmd,
            stdin=f,
            stderr=STDOUT, shell=True
        ).decode('utf8', 'replace')
    buffer = NamedStringIO()
    buffer.name = path
    buffer.write(output)
    buffer.seek(0)
    return buffer



#!/usr/bin/env python
# -*- coding:utf-8 -*-



import json
import os

from jinja2 import ChoiceLoader, FileSystemLoader

from jupyter_server.utils import url_path_join, to_os_path, ensure_async

generic_checkpoint_mixin_types = []
file_checkpoint_mixin_types = []

try:
    from jupyter_server.services.contents.checkpoints import GenericCheckpointsMixin as jpserver_GenericCheckpointsMixin
    from jupyter_server.services.contents.filecheckpoints import FileCheckpoints as jpserver_FileCheckpoints
    generic_checkpoint_mixin_types.append(jpserver_GenericCheckpointsMixin)
    file_checkpoint_mixin_types.append(jpserver_FileCheckpoints)
except ModuleNotFoundError:
    pass

try:
    from notebook.services.contents.checkpoints import GenericCheckpointsMixin as nbserver_GenericCheckpointsMixin
    from notebook.services.contents.filecheckpoints import FileCheckpoints as nbserver_FileCheckpoints
    generic_checkpoint_mixin_types.append(nbserver_GenericCheckpointsMixin)
    file_checkpoint_mixin_types.append(nbserver_FileCheckpoints)
except ModuleNotFoundError:
    pass

generic_checkpoint_mixin_types = tuple(generic_checkpoint_mixin_types)
file_checkpoint_mixin_types = tuple(file_checkpoint_mixin_types)


from tornado.web import HTTPError, escape, authenticated, gen

from ..args import process_diff_flags
from ..config import build_config, Namespace
from ..diffing.notebooks import set_notebook_diff_ignores, diff_notebooks
from ..gitfiles import (
    changed_notebooks, is_path_in_repo, find_repo_root,
    InvalidGitRepositoryError, BadName, GitCommandNotFound,
    GitRefWorkingTree, GitRefIndex
    )
from ..ignorables import diff_ignorables
from ..utils import read_notebook

from .nbdimeserver import (
    template_path,
    static_path,
    NbdimeHandler,
    MainDifftoolHandler,
    ApiDiffHandler,
    APIHandler,
)


special_refs = {
    'working': GitRefWorkingTree,
    'index': GitRefIndex,
}


class AuthMainDifftoolHandler(MainDifftoolHandler):
    @authenticated
    def get(self):
        return super(AuthMainDifftoolHandler, self).get()


class GitDifftoolHandler(NbdimeHandler):
    """Diff tool handler that also handles showing diff to git HEAD"""

    @authenticated
    def get(self):
        args = self.base_args()
        args['base'] = 'git:' + self.get_argument('base', '')
        args['remote'] = ''

        self.write(self.render_template(
            'difftool.html',
            config_data=args,
            ))


class CheckpointDifftoolHandler(NbdimeHandler):
    """Diff tool handler that also handles showing diff to git HEAD"""

    @authenticated
    def get(self):
        args = self.base_args()
        args['base'] = 'checkpoint:' + self.get_argument('base', '')
        args['remote'] = ''

        self.write(self.render_template(
            'difftool.html',
            config_data=args,
            ))


class BaseGitDiffHandler(ApiDiffHandler):

    def get_git_notebooks(self, file_path_arg, ref_base='HEAD', ref_remote=None):
        """
        Gets the content of the before and after state of the notebook based on the given Git refs.

        :param file_path_arg: The path to the file being diffed
        :param ref_base: the Git ref for the "local" or the "previous" state
        :param ref_remote: the Git ref for the "remote" or the "current" state
        :return: (base_nb, remote_nb)
        """
        # Sometimes the root dir of the files is not cwd
        nb_root = getattr(self.contents_manager, 'root_dir', None)
        # Resolve base argument to a file system path
        file_path = os.path.realpath(to_os_path(file_path_arg, nb_root))

        # Ensure path/root_dir that can be sent to git:
        try:
            git_root = find_repo_root(file_path)
        except InvalidGitRepositoryError as e:
            self.log.exception(e)
            raise HTTPError(422, 'Invalid notebook: %s' % file_path)
        file_path = os.path.relpath(file_path, git_root)

        # Get the base/remote notebooks:
        try:
            for fbase, fremote in changed_notebooks(ref_base, ref_remote, file_path, git_root):
                base_nb = read_notebook(fbase, on_null='minimal')
                remote_nb = read_notebook(fremote, on_null='minimal')
                break  # there should only ever be one set of files
            else:
                # The filename was either invalid or the file is unchanged
                # Assume unchanged, and let read_notebook handle error
                # reporting if invalid
                base_nb = self.read_notebook(os.path.join(git_root, file_path))
                remote_nb = base_nb
        except (InvalidGitRepositoryError, BadName) as e:
            self.log.exception(e)
            raise HTTPError(422, 'Invalid notebook: %s' % file_path_arg)
        except GitCommandNotFound as e:
            self.log.exception(e)
            raise HTTPError(
                500, 'Could not find git executable. '
                     'Please ensure git is available to the server process.')

        return base_nb, remote_nb

    @property
    def curdir(self):
        root_dir = getattr(self.contents_manager, 'root_dir', None)
        if root_dir is None:
            return super(ExtensionApiDiffHandler, self).curdir
        return root_dir


class ExtensionApiDiffHandler(BaseGitDiffHandler):
    """Diff API handler that also handles diff to git HEAD"""

    async def _get_checkpoint_notebooks(self, base):
        # Get the model for the current notebook:
        cm = self.contents_manager
        model = await ensure_async(cm.get(base, content=True, type='notebook'))
        remote_nb = model['content']
        # Get the model for the checkpoint notebook:
        checkpoints = await ensure_async(cm.list_checkpoints(base))
        if not checkpoints:
            # No checkpoints, indicate unchanged:
            self.log.info('No checkpoints for file: %r, %r', base, checkpoints)
            return remote_nb, remote_nb
        self.log.debug('Checkpoints: %r', checkpoints)
        checkpoint = checkpoints[0]
        if isinstance(cm.checkpoints, generic_checkpoint_mixin_types):
            checkpoint_model = await ensure_async(
                cm.checkpoints.get_notebook_checkpoint(checkpoint, base))
            base_nb = checkpoint_model['content']
        elif isinstance(cm.checkpoints, file_checkpoint_mixin_types):
            path = await ensure_async(
                cm.checkpoints.checkpoint_path(checkpoint['id'], base))
            base_nb = read_notebook(path, on_null='minimal')
        else:
            raise RuntimeError('Unknown checkpoint handler interface')
        return base_nb, remote_nb

    @authenticated
    async def post(self):
        # TODO: Add deprecation warning (for git/checkpoint only?)
        # Assuming a request on the form "{'argname':arg}"
        body = json.loads(escape.to_unicode(self.request.body))
        base = body['base']
        if base.startswith('git:'):
            base_nb, remote_nb = self.get_git_notebooks(base[len('git:'):])
        elif base.startswith('checkpoint:'):
            base_nb, remote_nb = await self._get_checkpoint_notebooks(base[len('checkpoint:'):])
        else:
            # Regular files, call super
            super(ExtensionApiDiffHandler, self).post()
            return

        # Perform actual diff and return data:
        try:
            thediff = diff_notebooks(base_nb, remote_nb)
        except Exception:
            self.log.exception('Error diffing documents:')
            raise HTTPError(500, 'Error while attempting to diff documents')

        data = {
            'base': base_nb,
            'diff': thediff,
            }
        self.finish(data)


class GitDiffHandler(BaseGitDiffHandler):
    """Diff API handler that handles diffs for two git refs"""

    @classmethod
    def parse_ref(cls, data):
        return data.get('git', None) or special_refs[data['special'].lower()]

    def _validate_request(self, body):
        def _fail(msg):
            self.log.exception(msg)
            raise HTTPError(400, msg)

        # Validate refs
        for refname in ('ref_local', 'ref_remote'):

            # Validate ref_curr
            try:
                ref = body[refname]
            except KeyError:
                _fail('Required key %s not provided in the request' % (refname))

            # Either of special or git is supported in ref
            if 'special' in ref and 'git' in ref:
                _fail('Only one of special and git should be present in git '
                      'reference.')

            if not ('special' in ref or 'git' in ref):
                _fail('At least one of special and git should be present in git '
                      'reference.')

            if 'special' in ref:
                special = ref['special'].lower()
                if refname == 'ref_local':
                    if special != 'index':
                        _fail('Only "index" is allowed for the "special" value '
                              'on ref_local, got %r.' % (special,))
                elif special not in ('index', 'working'):
                    _fail('Only "index" or "working" is allowed for the "special" value '
                          'on ref_remote, got %r.' % (special,))


        # Validate file_name
        try:
            body['file_path']
        except KeyError:
            _fail('Required value file_path not provided in the request')


    @authenticated
    @gen.coroutine
    def post(self):
        body = json.loads(escape.to_unicode(self.request.body))

        try:
            # Validate the request input
            self._validate_request(body)

            # Get file contents based on Git regs
            ref_local = body['ref_local']
            ref_remote = body['ref_remote']
            file_path = body['file_path']
            base_nb, remote_nb = self.get_git_notebooks(
                file_path,
                GitDiffHandler.parse_ref(ref_local),
                GitDiffHandler.parse_ref(ref_remote),
            )

            # Perform actual diff and return data
            thediff = diff_notebooks(base_nb, remote_nb)

            data = {
                'base': base_nb,
                'diff': thediff,
            }
            self.finish(data)
        except HTTPError:
            raise
        except Exception:
            self.log.exception('Error diffing documents:')
            raise HTTPError(500, 'Error while attempting to diff documents')


class IsGitHandler(NbdimeHandler, APIHandler):
    """API handler for querying if path is in git repo"""

    @authenticated
    def post(self):
        root_dir = getattr(self.contents_manager, 'root_dir', None)
        # Ensure notebooks are file-system based
        if root_dir is None:
            self.finish({'is_git': False})

        # Assuming a request on the form "{'argname':arg}"
        body = json.loads(escape.to_unicode(self.request.body))
        nb = os.path.join(root_dir, body['path'])

        data = {'is_git': is_path_in_repo(nb)}
        self.finish(data)


def _load_jupyter_server_extension(nb_server_app, nb6_entrypoint=False):
    """
    Called when the extension is loaded.

    Args:
        nb_server_app (NotebookWebApplication): handle to the Notebook webserver instance.
    """
    if nb6_entrypoint:
        # We're using the old notebook 6 extenstion entry point
        # In this case, we only support jupyter_server < 2, so fail if >=2
        from jupyter_server._version import __version__
        try:
            from packaging.version import parse, Version
            if parse(__version__) >= Version('2.0.0'):
                nb_server_app.log.critical(
                    "You must use Jupyter Server v1 to load nbdime as a classic notebook server extension. "
                    f"You have v{__version__} installed.\nYou can fix this by executing:\n"
                    "    pip install -U \"jupyter-server<2.0.0\""
                )
                return
        except Exception: # noqa
            pass
    web_app = nb_server_app.web_app

    env = web_app.settings['jinja2_env']

    env.loader = ChoiceLoader([
        env.loader,
        FileSystemLoader(template_path),
    ])

    config = build_config('extension')
    ignore = config.pop('Ignore', None)
    for k in diff_ignorables:
        config[k] = config.get(k, None)
    ns = Namespace(config)
    process_diff_flags(ns)
    if ignore:
        set_notebook_diff_ignores(ignore)

    web_app.settings['static_path'].append(static_path)

    params = {
        'nbdime_relative_base_url': 'nbdime',
        'closable': False,
    }
    handlers = [
        (r'/nbdime/difftool', AuthMainDifftoolHandler, params),
        (r'/nbdime/checkpoint-difftool', CheckpointDifftoolHandler, params),
        (r'/nbdime/git-difftool', GitDifftoolHandler, params),
        (r'/nbdime/api/diff', ExtensionApiDiffHandler, params),
        (r'/nbdime/api/isgit', IsGitHandler, params),
        (r'/nbdime/api/gitdiff', GitDiffHandler, params)
    ]

    # Prefix routes with base_url:
    base_url = web_app.settings.get('base_url', '/')
    handlers = [(url_path_join(base_url, h[0]), h[1], h[2]) for h in handlers]

    host_pattern = '.*$'
    web_app.add_handlers(host_pattern, handlers)



#!/usr/bin/env python
# -*- coding:utf-8 -*-




import sys

import warnings

from .nbdimeserver import main_server as run_server
from .webutil import browse as browse_util
from ..args import (
    ConfigBackedParser,
    add_generic_args, add_web_args, add_diff_args,
    args_for_server, args_for_browse, process_diff_flags,
    resolve_diff_args)
from ..gitfiles import changed_notebooks, is_gitref


def build_arg_parser():
    """
    Creates an argument parser for the diff tool, that also lets the
    user specify a port and displays a help message.
    """
    description = 'Difftool for Nbdime.'
    parser = ConfigBackedParser(
        description=description,
        add_help=True
        )
    add_generic_args(parser)
    add_web_args(parser, 0)
    add_diff_args(parser)
    parser.add_argument(
        "base", help="The base notebook filename OR base git-revision.",
        nargs='?', default='HEAD',
    )
    parser.add_argument(
        "remote", help="The remote modified notebook filename OR remote git-revision.",
        nargs='?', default=None,
    )
    parser.add_argument(
        "paths", help="Filter diffs for git-revisions based on path",
        nargs='*', default=None,
    )
    return parser


def browse(port, base, remote, browser):
    browse_util(port=port,
                browsername=browser,
                rel_url='diff',
                base=base,
                remote=remote)
    warnings.warn(
        'This function is deprecated. '
        'Use nbdime.webapp.webutil.browse() instead.',
        DeprecationWarning)


def handle_gitrefs(base, remote, path, arguments):
    status = 0
    for fbase, fremote in changed_notebooks(base, remote, path):
        status = run_server(
            difftool_args=dict(base=fbase, remote=fremote),
            on_port=lambda port: browse_util(
                port=port,
                rel_url='difftool',
                **args_for_browse(arguments)),
            **args_for_server(arguments))
        if status != 0:
            # Short-circuit for errors along the way
            return status
    return status


def main_diff(opts):
    process_diff_flags(opts)
    base, remote, path = resolve_diff_args(opts)
    if is_gitref(base) and is_gitref(remote):
        # We are asked to do a gui for git diff
        return handle_gitrefs(base, remote, path, opts)
    return run_server(
        on_port=lambda port: browse_util(
            port=port,
            rel_url='diff',
            base=base, remote=remote,
            **args_for_browse(opts)),
        **args_for_server(opts))


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    opts = build_arg_parser().parse_args(args)
    return main_diff(opts)


if __name__ == "__main__":
    sys.exit(main())




import os
import shutil
from itertools import zip_longest

import pytest

from git import InvalidGitRepositoryError

from ..gitfiles import changed_notebooks
from ..utils import EXPLICIT_MISSING_FILE


# Test that it can diff
#  - one ref
#  - two refs
#  - all of the above with a directory path
#  - all of the above with a filename path
#  - all of the above within a subdirectory of the repo
#  - all of the above with added and removed files in diff
#  - does not attempt to diff non-Notebook files
# Fails if:
#  - path is non-sensical
#  - path is not a git repo
#  - one or both references are invalid
#  - base ref is None


def _nb_name(f):
    if isinstance(f, str):
        return f
    return f.name


_filename = 'diff.ipynb'
_subdir_name = 'sub'
_subdir_filename = 'subfile.ipynb'
_subpath = os.path.join(_subdir_name, _subdir_filename)


# Test one/two args:

def test_head_vs_workdir(git_repo2):
    expected = [
        ('diff.ipynb (HEAD)', EXPLICIT_MISSING_FILE),
    ]
    for expected, actual in zip_longest(expected, changed_notebooks('HEAD', None), fillvalue=None):
        assert _nb_name(actual[0]) == expected[0]
        assert _nb_name(actual[1]) == expected[1]


def test_ref_vs_ref(git_repo2):
    expected = [
        ('diff.ipynb (base)', 'diff.ipynb (local)'),
        ('sub/subfile.ipynb (base)', 'sub/subfile.ipynb (local)'),
    ]
    for expected, actual in zip_longest(expected, changed_notebooks('base', 'local'), fillvalue=None):
        assert _nb_name(actual[0]) == expected[0]
        assert _nb_name(actual[1]) == expected[1]


# Test one/two args executed in subdir. Subdir should have no effect without `path` arg:

def test_head_vs_workdir_subdir(git_repo2):
    os.chdir(os.path.join(git_repo2, _subdir_name))
    expected = [
        ('diff.ipynb (HEAD)', EXPLICIT_MISSING_FILE),
    ]
    for expected, actual in zip_longest(expected, changed_notebooks('HEAD', None), fillvalue=None):
        assert _nb_name(actual[0]) == expected[0]
        assert _nb_name(actual[1]) == expected[1]


def test_ref_vs_ref_subdir(git_repo2):
    os.chdir(os.path.join(git_repo2, _subdir_name))
    expected = [
        ('diff.ipynb (base)', 'diff.ipynb (local)'),
        ('sub/subfile.ipynb (base)', 'sub/subfile.ipynb (local)'),
    ]
    for expected, actual in zip_longest(expected, changed_notebooks('base', 'local'), fillvalue=None):
        assert _nb_name(actual[0]) == expected[0]
        assert _nb_name(actual[1]) == expected[1]


# Test one/two args executed in subdir, with subdir relative path:

def test_head_vs_workdir_subdir_with_path(git_repo2, filespath):
    # Ensure diff for subfile:
    full_subfile_path = os.path.join(git_repo2, _subdir_name, 'subfile.ipynb')
    shutil.copy(os.path.join(filespath, 'foo--1.ipynb'),
                full_subfile_path)
    os.chdir(os.path.join(git_repo2, _subdir_name))
    expected = [
        ('sub/subfile.ipynb (HEAD)', 'sub/subfile.ipynb'),
    ]
    for expected, actual in zip_longest(expected, changed_notebooks('HEAD', None, _subdir_filename), fillvalue=None):
        assert _nb_name(actual[0]) == expected[0]
        assert _nb_name(actual[1]) == expected[1]


def test_ref_vs_ref_subdir_with_path(git_repo2):
    os.chdir(os.path.join(git_repo2, _subdir_name))
    expected = [
        ('sub/subfile.ipynb (base)', 'sub/subfile.ipynb (local)'),
    ]
    for expected, actual in zip_longest(expected, changed_notebooks('base', 'local', _subdir_filename), fillvalue=None):
        assert _nb_name(actual[0]) == expected[0]
        assert _nb_name(actual[1]) == expected[1]



# Test one/two args with subdir path:

def test_head_vs_workdir_subdir_path(git_repo2):
    expected = [
    ]
    for expected, actual in zip_longest(expected, changed_notebooks('HEAD', None, _subdir_name), fillvalue=None):
        assert _nb_name(actual[0]) == expected[0]
        assert _nb_name(actual[1]) == expected[1]


def test_ref_vs_ref_subdir_path(git_repo2):
    expected = [
        ('sub/subfile.ipynb (base)', 'sub/subfile.ipynb (local)'),
    ]
    for expected, actual in zip_longest(expected, changed_notebooks('base', 'local', _subdir_name), fillvalue=None):
        assert _nb_name(actual[0]) == expected[0]
        assert _nb_name(actual[1]) == expected[1]


# Test one/two args with filename path:

def test_head_vs_workdir_filename(git_repo2):
    expected = [
        ('diff.ipynb (HEAD)', EXPLICIT_MISSING_FILE),
    ]
    for expected, actual in zip_longest(expected, changed_notebooks('HEAD', None, _filename), fillvalue=None):
        assert _nb_name(actual[0]) == expected[0]
        assert _nb_name(actual[1]) == expected[1]


def test_ref_vs_ref_filename(git_repo2):
    expected = [
        ('diff.ipynb (base)', 'diff.ipynb (local)'),
    ]
    for expected, actual in zip_longest(expected, changed_notebooks('base', 'local', _filename), fillvalue=None):
        assert _nb_name(actual[0]) == expected[0]
        assert _nb_name(actual[1]) == expected[1]


# Test one/two args executed in subdir with filename path:

def test_head_vs_workdir_subdir_filename(git_repo2):
    expected = [
    ]
    os.chdir(os.path.join(git_repo2, _subdir_name))
    for expected, actual in zip_longest(expected, changed_notebooks('HEAD', None, _subdir_filename), fillvalue=None):
        assert _nb_name(actual[0]) == expected[0]
        assert _nb_name(actual[1]) == expected[1]


def test_ref_vs_ref_subdir_filename(git_repo2):
    expected = [
        ('sub/subfile.ipynb (base)', 'sub/subfile.ipynb (local)'),
    ]
    os.chdir(os.path.join(git_repo2, _subdir_name))
    for expected, actual in zip_longest(expected, changed_notebooks('base', 'local', _subdir_filename), fillvalue=None):
        assert _nb_name(actual[0]) == expected[0]
        assert _nb_name(actual[1]) == expected[1]


# Test one/two args with path to file in subdir:

def test_head_vs_workdir_subdir_filename_path(git_repo2):
    expected = [
    ]
    for expected, actual in zip_longest(expected, changed_notebooks('HEAD', None, _subpath), fillvalue=None):
        assert _nb_name(actual[0]) == expected[0]
        assert _nb_name(actual[1]) == expected[1]


def test_ref_vs_ref_subdir_filename_path(git_repo2):
    expected = [
        ('sub/subfile.ipynb (base)', 'sub/subfile.ipynb (local)'),
    ]
    for expected, actual in zip_longest(expected, changed_notebooks('base', 'local', _subpath), fillvalue=None):
        assert _nb_name(actual[0]) == expected[0]
        assert _nb_name(actual[1]) == expected[1]


# Test failure of one/two args with path to invalid file:

def test_head_vs_workdir_non_existant(git_repo2):
    assert 0 == len(tuple(changed_notebooks('HEAD', None, 'non-existant-file.ipynb')))


def test_ref_vs_ref_non_existant(git_repo2):
    assert 0 == len(tuple(changed_notebooks('base', 'local', 'non-existant-file.ipynb')))


# Test failure of one/two args with (invalid) path to file in parent directory:

def test_head_vs_workdir_invalid_path(git_repo2):
    os.chdir(os.path.join(git_repo2, _subdir_name))
    assert 0 == len(tuple(changed_notebooks('HEAD', None, _filename)))


def test_ref_vs_ref_invalid_path(git_repo2):
    os.chdir(os.path.join(git_repo2, _subdir_name))
    assert 0 == len(tuple(changed_notebooks('base', 'local', _filename)))


# Test failure for invalid directory path
def test_head_vs_workdir_invalid_subdir(git_repo2):
    assert 0 == len(tuple(changed_notebooks('HEAD', None, 'non-existant-dir')))


# Test failure outside git repo
def test_no_repo(tmpdir):
    tmpdir.chdir()
    with pytest.raises(InvalidGitRepositoryError):
        tuple(changed_notebooks('HEAD', None))



# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

"""This file contains tests applying to reference notebook files from the nbdime/tests/files/ directory."""




import nbformat

from nbdime import patch, patch_notebook, diff_notebooks
from nbdime.diffing.notebooks import diff_cells

# pytest conf.py stuff is tricky to use robustly, this works with no magic
from .utils import assert_is_valid_notebook, check_diff_and_patch


def test_notebook_database_fixture(db):
    "Just test that the notebook file reader fixture is at least self-consistent."
    assert len(db) > 0
    assert len(list(db.keys())) == len(db)
    for name in db:
        assert name in db
        assert name + "notindb" not in db
        assert_is_valid_notebook(db[name])


def test_diff_and_patch_metadata_of_notebooks(any_nb_pair):
    "Test diff/patch on the metadata of any pair of notebooks in the test suite."
    nba, nbb = any_nb_pair
    a = nba["metadata"]
    b = nbb["metadata"]
    check_diff_and_patch(a, b)


def test_diff_and_patch_notebooks_with_generic_diff(any_nb_pair):
    "Test generic diff/patch on any pair of notebooks in the test suite."
    a, b = any_nb_pair
    check_diff_and_patch(a, b)


def test_diff_and_patch_cells_of_notebooks(any_nb_pair):
    "Test diff/patch on the cells of any pair of notebooks in the test suite."
    nba, nbb = any_nb_pair
    a = nba["cells"]
    b = nbb["cells"]
    assert patch(a, diff_cells(a, b)) == b


def test_diff_and_patch_matching_notebooks(matching_nb_pairs):
    "Test diff/patch on pairs of notebooks with the same basename in the test suite."
    a, b = matching_nb_pairs
    assert patch_notebook(a, diff_notebooks(a, b)) == nbformat.from_dict(b)


def test_diff_and_patch_notebooks(any_nb_pair):
    "Test diff/patch on any pair of notebooks in the test suite."
    a, b = any_nb_pair
    assert patch_notebook(a, diff_notebooks(a, b)) == nbformat.from_dict(b)
